package org.icroco.tablemodel.blinking; // Generated package name

import java.util.Queue;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.Delayed;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicLong;

/**
 * DataChange is used to associate a Data Object with a column identifier that
 * has changed. To avoid loads of garbage per update DataChanges are cached and
 * reused.
 */
public final class DataChange
    implements Delayed
{
    /**
     * DEFAULT_EXPIRATION_TIME_IN_MS
     */
    public final static long         DEFAULT_EXPIRATION_TIME_IN_MS = 1000;
    /** Base of nanosecond timings, to avoid wrapping */
    private static final AtomicLong  sequencer                     = new AtomicLong(0);

    private final long               sequenceNumber                = sequencer.getAndIncrement();

    private long                     expiration                    = DEFAULT_EXPIRATION_TIME_IN_MS;

    private long                     time                          = this.now();

    private static Queue<DataChange> sharedDataChanges             = new ArrayBlockingQueue<DataChange>(2000);

    private int                      col;
    private int                      row;
    private transient int            hashcode;

    protected long getExpiration()
    {
        return this.expiration;
    }

    protected final void setExpiration(final long expirationInMs)
    {
        this.expiration = TimeUnit.NANOSECONDS.convert(expirationInMs, TimeUnit.MILLISECONDS);
        this.time = this.now();
    }

    /**
     * Returns nanosecond time offset by origin
     */
    final long now()
    {
        return System.nanoTime();
    }

    void resetTime()
    {
        this.time = this.now();
    }

    /**
     * Obtains a DataChange for the specified Data and column.
     * 
     * @param row
     * @param col
     * @param exp
     * @return
     */
    public static final DataChange getDataChange(final int row, final int col, final long exp)
    {
        final DataChange change = getDataChange(row, col);
        change.setExpiration(exp);

        return change;
    }

    /**
     * getDataChange.
     * 
     * @param row
     * @param col
     * @return
     */
    public static final DataChange getDataChange(final int row, final int col)
    {
        DataChange change = sharedDataChanges.poll();
        if (change != null)
        {
            change.row = row;
            change.col = col;
            change.resetHashcode();
        } else
        {
            change = new DataChange(row, col);
        }

        return change;
    }

    /**
     * Indicates the DataChange is no longer needed and can be reused.
     * 
     * @param change
     */
    public static void releaseDataChange(final DataChange change)
    {
        sharedDataChanges.add(change);
    }

    /**
     * size.
     * 
     * @return
     */
    public static int size()
    {
        return sharedDataChanges.size();
    }

    DataChange(final int aRow, final int aCol, final long exp)
    {
        this(aRow, aCol);
        this.setExpiration(exp);
    }

    DataChange(final int aRow, final int aCol)
    {
        this.row = aRow;
        this.col = aCol;
        this.resetHashcode();
    }

    /**
     * getColumn.
     * 
     * @return
     */
    public int getColumn()
    {
        return this.col;
    }

    /**
     * @return Returns the row.
     */
    public int getRow()
    {
        return this.row;
    }

    /**
     * Override hashCode.
     * 
     * @return the Objects hashcode.
     */
    @Override
    public int hashCode()
    {
        // means collision at grid > 10000 row
        // return (row+10000*col);
        return this.hashcode;
    }

    final void resetHashcode()
    {
        this.time = this.now();
        this.hashcode = (this.row ^ (this.row >>> 32)) & (this.col ^ (this.col >>> 32));
    }

    /**
     * (non-Javadoc)
     * 
     * @see java.lang.Object#equals(java.lang.Object)
     */
    @Override
    public boolean equals(final Object dc)
    {
        if (dc == this)
        {
            return true;
        }

        final DataChange o = (DataChange) dc;
        return (o.row == this.row && o.col == this.col);
    }

    /**
     * (non-Javadoc)
     * 
     * @see java.util.concurrent.Delayed#getDelay(java.util.concurrent.TimeUnit)
     */
    public long getDelay(final TimeUnit unit)
    {
        // long now = now();
        // System.out.println("now = "+now);
        // System.out.println("time = "+time);
        // System.out.println("expiration = "+expiration);
        return this.time + this.expiration - this.now();
    }

    /**
     * (non-Javadoc)
     * 
     * @param other
     * @return
     * 
     * @see java.lang.Comparable#compareTo(T)
     */
    public int compareTo(final Delayed other)
    {
        if (other == this)
        {
            return 0;
        }
        final DataChange x = (DataChange) other;
        final long diff = this.time - x.time;
        if (diff < 0)
        {
            return -1;
        } else if (diff > 0)
        {
            return 1;
        } else if (this.sequenceNumber < x.sequenceNumber)
        {
            return -1;
        } else
        {
            return 1;
        }
    }

    /**
     * @param aCol
     *            the col to set
     */
    void setCol(final int aCol)
    {
        this.col = aCol;
    }

    /**
     * @param aRow
     *            the row to set
     */
    void setRow(final int aRow)
    {
        this.row = aRow;
    }

    /**
     * (non-Javadoc)
     * 
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString()
    {
        final StringBuffer buffer = new StringBuffer();
        buffer.append("[DataChange:");
        buffer.append(" row: ");
        buffer.append(this.row);
        buffer.append(" col: ");
        buffer.append(this.col);
        buffer.append("]");
        return buffer.toString();
    }

}
